/* debug.c - a few functions to help with debugging programs
	out("some text") - sprinkle a few of these thru the
	program to locate segfaults. Quicker than gdb for that.
	track("some text") - appends to a file 'track' to record
	the state of char * variables. Opens, writes, flushes 
	and closes the file every call.
	trackmem(check) - appends 10 bytes beginning at check plus
	'\n' to file 'track'. Operates same as track().
	killtrack() - invoke it when the program starts up to
	remove file 'track' if it exists. Needed because the file 
	is oppened in mode "a", not "w".
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/stat.h>
#include "debug.h"

void out(char *what) {
	printf("%s\n", what);
	fflush(stdout);
} // out()
void track(char *what) {
	FILE *fpo = fopen("track", "a");
	fprintf(fpo, "%s\n", what);
	fflush(fpo);
	fclose(fpo);
	return;
} //track()
void trackmem(void *check) {
	int i;
	char *cp;
	FILE *fpo = fopen("track", "a");
	for (i=0; i<10; i++) {
		cp = check + i;
		fputc(*cp, fpo);
	}
	fprintf(fpo, "\n");
	fflush(fpo);
	fclose(fpo);
	return;
}// trackmem()
void killtrack(void){
	int c;
	STAT st;
	c = stat("track", &st);
	if (c == 0)
		c = unlink("track");
	return;
}// killtrack()
void trackp(void *p){
	FILE *fpo = fopen("track", "a");
	fprintf(fpo, "%p\n", p);
	fflush(fpo);
	fclose(fpo);
	return;
	
}
