/* ingest_file.c - function to ingest a file (surprise!)
	stats the file, finds the size, allocates a buffer
	of that size. All being well returns a pointer to
	a struct carrying void pointers to start and end
	of the buffer. End pointer points to byte after
	the last in buffer. Checks that file ends with a
	newline, if not, it appends it.
	Aborts with error message if anything goes wrong. 
*/
#include<stdio.h>
#include<string.h>
#include<stdlib.h>
#include<ctype.h>
#include<sys/stat.h>
#include<errno.h>
#include"ingest_file.h"

IFRETURN *
ingest_file(char* path) {
	int c;
	char *cp;
	size_t count;
	FILE *fpi;
	char error_msg[256];
	void *start, *end;
	STAT *st = malloc(sizeof(STAT));
	IFRETURN *ifr = malloc(sizeof(IFRETURN));
	if (!(st && ifr)) {
		fatal_error("no memory for structs in ingest_file()");
	}
	if ((c = stat(path, st))) {
		snprintf(error_msg, 255,
			"Could not stat file: %s, in ingest_file()",
			 path);
		fatal_error(error_msg);
	}
	if (st->st_size == 0) {
		errno = 127;	// "unknown error"
		snprintf(error_msg, 255,
			 "Zero size file: %s, in ingest_file()",
			 path);
		fatal_error(error_msg);
	}
	fpi = fopen(path, "r");
	if (!(fpi)) {
		snprintf(error_msg, 255,
			 "Could not open file: %s, in ingest_file()",
			 path);
		fatal_error(error_msg);
	}
	start = malloc(st->st_size + 1); // in case no '\n' at end
	if (!(start)) {
		snprintf(error_msg, 255,
			 "Could not get memory %ld bytes, in ingest_file()",
			 st->st_size + 1);
		fatal_error(error_msg);
	}
	count = fread(start, sizeof(char), st->st_size, fpi);
	if ( count != st->st_size) {
		snprintf(error_msg, 255,
			 "File read error in ingest_file(), "
			"%ld expected, %d read",
			 st->st_size, count);
		fatal_error(error_msg);
	}
	end = start + st->st_size;
	// make sure there is a newline at the last byte
	cp = end - 1;
	if (*cp != '\n') {
            // I made enough room for this when 'malloc'ed the buffer
            cp++;
            *cp = '\n';
            end++;
	}// if (no '\n' at end)
	ifr->start = start;
	ifr->end = end;
	free(st);
	fclose(fpi);
	return ifr;
}// ingest_file()
void 
fatal_error(char *msg){
	perror(msg);
	fflush(stderr);
	exit(1);
}
