#ifndef _XLIST_H
#define _XLIST_H	1

 typedef struct ln {
	void *data;
	struct ln *next;
} LNODE;

LNODE *new_node(void);
/*  assign one such to be the list head, then nodes to insert as required.
    The data pointer of the head node remains NULL */

LNODE *ol_insert(LNODE *olnode, LNODE *newnode, 
			int (*comp)(void *, void *));
/*insert newnode to ordered list. Inserts where it finds data > itself.
    If no such data exists newnode is appended to the list */

LNODE *ul_append(LNODE *head, LNODE *newnode);
/* If you don't need ordered data, or if the data is already sorted,
    use this to append the newnode to the list. */

void destroy_node(LNODE *node, void (*destroy)(void *));
/* Frees up the node after deletion. The destroy() function frees the
    data and then free() is called for the node. For char* data free()
    does the destroy() job. */

void destroy_list(LNODE *head, void (*destroy)(void *));
/* Destroys the entire list including the head node */

void node_traverse(LNODE *node, int (*do_what)(void *));
/* Travels forward thru the entire list doing whatever do_what() does.
   To print char* data puts() does the job. */

void node_traverse_reverse(LNODE *node, int (*do_what)(void *));
/* Travels the list in the opposite order, for an ordered list you
    can get it in highest to lowest order */

void fatal_error(char *msg);
/* invokes perror() with msg and then exit(1) the program */

LNODE *ul_push(LNODE *head, LNODE *newnode);
/* Inserts newnode as the first item in the list ie between the head
    node and the rest of the list. Use for a stack LIFO */

LNODE *ul_pop(LNODE *head);
/* Links the head node to the list behind the first node. Returns that
    first node to the caller which is responsible for freeing it. */

LNODE *ol_find(LNODE *olnode, void *what, int (*comp)(void *, void *));
/* Finds the first node that matches 'what' and returns the pointer to it
 if any. Returns NULL if it finds data < what or reaches end of list. */

LNODE *ol_find_before(LNODE *olnode, void *what, int (*comp)(void *, void *));
/* Returns the node before the one with matching data if any. Used by 
    ol_delete() to prepare for deletion. Returns NULL if not found. */

LNODE *ul_find_before(LNODE *ulnode, void *what, int (*comp)(void *, void *));
/* Returns the node before the match if any. Unlike its ol counterpart it
    traverses the entire list if there is no match. ul_delete() uses it.*/

LNODE *ul_find(LNODE *ulnode, void *what, int (*comp)(void *, void *));
/* Returns the node that matches if any. Unlike its ol counterpart it
    traverses the entire list if there is no match. NULL if not found. */

LNODE *ul_delete(LNODE *ulnode, void *what,
		int (*comp)(void *, void *), void (*destructor) (void));
/* Delete if 'what' is found and return pointer node before, otherwise
    return NULL*/

LNODE *ol_delete(LNODE *olnode, void *what, 
		int (*comp)(void *, void *), void (*destructor) (void));
/* Same as ul_delete except that 'not found' happens if data item is
    less than 'what' as well as end of list reached. */

LNODE *node_delete(LNODE *node, LNODE *deleteit, void (*data_destroy) (void));
/* Delete when node is known. Returns pointer to node before or NULL if
    node does not exist */

#endif /* xlist.h */
