/* when.c - a program to produce lists of the nature:
   "first and third Tuesday in the month", "last Friday in the month",
    and so on.

*/
/*
struct tm
{
	int tm_sec;	// Seconds	[0-60]
	int tm_min;	// Minutes	[0-59]
	int tm_hour;	// Hours	[0-23]
	int tm_mday;	// Day in Month	[1-31]
	int tm_mon;	// Month	[0-11]
	int tm_year;	// Year - 1900
	int tm_wday;	// Day in Week	[0..6]
	int tm_yday;	// Day in Year	[0-365]
	int tm_isdst;	// Daylight S T	[-1/0/1]

};
*/

#include<stdio.h>
#include<string.h>
#include<stdlib.h>
#include<ctype.h>
#include <unistd.h>
#include <time.h>
#include "xlist.h"

typedef struct tm TIME;
typedef struct hm {
    int hh;         // [0..24]
    int mm;         // [0..59]
    int amflag;     // user spec'd am/pm time [0..1]
    int invalid;    // [0..1]
} HM;
typedef struct dd { // 
    int yy;         // year - as ccyy
    int mon;        // month
    int dd;         // day
    int hh;         // hour
    int mm;         // minute
    int wd;         // week day [0..6]
    int pim_mask;   // 1st = 1, 2nd = 2, 3rd = 4
                    // 4th = 8, 5th or last = 16
 // and so the 4th when it's also the last will be 8|16
} DATEDATA;

// function prototypes
void help(int err_status);
DATEDATA *new_dd(TIME *tv);
TIME *get_now(void);
TIME *set_dom(TIME *tp, int dom);
void arglist_print(int argc, char **argv);
char *propercase(char *dayname);
int weekday_number(char *dayname, char *ddnames[]);
int validate_dim(char *dim);
int validate_month(char *mo, char **monames);
int comp(DATEDATA *left, DATEDATA *right);
void print_node(DATEDATA *tp, char **ddnames, char **mmnames, int reqtime,
                int mergeable, int ampm, int trunc_dm, int dbg);
HM *validate_time(const char *ts);
time_t set_diw(int diw, time_t tim);
void set_pim(LNODE *node);
int get_month_next_week(DATEDATA *dd);		
void delete_nodes_unwanted(LNODE *node, int yy, int mon,
                           int nm, int wanted_mask, int except_mon);

int main (int argc, char** argv) {
    TIME *tp;
    time_t tnow, tend, tstart;
    int c;
    int yy, mon;        //year, month
    char dim[10];       // day in month [1..5] where 5 == 'last'
    int diw;            // day in week [0..6]; (Sun..Sat)
    int nm;             // number of months
    int em;             // except this month [1..12]
    DATEDATA *dd = NULL;
    LNODE *head = new_node();
    LNODE *test = new_node();
    LNODE *newnode = NULL;
    int wanted_mask = 0;
    int hh, mm, print_time; // user specified time if any
    HM *hm;                 // ditto
    int amflag;             // 24 hr | am pm
    char *headline;         // report header
    int inh_header;         // inhibit printing header
    int mergeable;          // output a leading "ccyymmddhhmm\t"
    // so that 2 or more runs may be meaningfully sorted eg meetings
    // happen on 2nd Friday at 7pm and 3rd Thursday at 3 pm
    int trunc_dm;           // truncate day & month names to 3 chars
    char *mmnames[] = {"error",
                    "January",
                    "February",
                    "March",
                    "April",
                    "May",
                    "June",
                    "July",
                    "August",
                    "September",
                    "October",
                    "November",
                    "December"
    };
    char *ddnames[] = {"Sunday",
                    "Monday",
                    "Tuesday",
                    "Wednesday",
                    "Thursday",
                    "Friday",
                    "Saturday"
    };
    // set up our defaults, to be overwritten by our args if required
    tp = get_now();     // current time
    tp = set_dom(tp, 1);// back to 1st day
    diw = -1;           // user must specify this
    nm = 3;		// user MAY specify this
    strcpy(dim, "-1");  // user must specify this
    em = 0;             // user MAY specify this
    hh = 12;            // default time values
    mm = 0;
    print_time = 0;     // don't output time unless user inputs it
    amflag = 0;         // default is 24 hr time
    headline = strdup("Events");
    inh_header = 0;     // default is to print the header
    mergeable = 0;      // default is not to print sortfield
    trunc_dm = 0;       // default is print long day & month names
    while ((c = getopt(argc, argv, ":hs:n:d:w:e:c:l:imt")) != -1) {
        switch(c) {
            case 'h':
                help(0);
                break;
            case 's':
            { //oa
                    char *oa = strdup(optarg);
                    if (strlen(oa) == 6) {
                        mon = atoi(&oa[4]);
                        oa[4] = '\0';
                        yy = atoi(oa);
                    } else if (strlen(oa) == 4) {
                        mon = atoi(&oa[2]);
                        oa[2] = '\0';
                        yy = 2000 + atoi(oa);
                    } else {
                        fputs("Wrong length for start date\n", stdout);
                        arglist_print(argc, argv);
                        help(1);
                    }
                    // we have our tp already intitialised
                    tp->tm_year = yy - 1900;
                    tp->tm_mon	= mon - 1;
                    free(oa);
                } // char(oa)
                break;
            case 'n':
                nm = atoi(optarg);
                if (nm < 1) {
                    printf("Months must be 1 or more\n");
                    arglist_print(argc, argv);
                    help(1);
                }
                break;
            case 'd':
                if ((diw = weekday_number(optarg, ddnames)) < 0) {
                    printf("Invalid weekday argument\n");
                    arglist_print(argc, argv);
                    help(1);
                }
                break;
            case 'w':
                strcpy(dim, optarg);
                if (validate_dim(dim)) {
                    printf("Invalid day(s) in month argument\n");
                    arglist_print(argc, argv);
                    help(1);
                }
                break;
            case 'e':
                em = atoi(optarg);
                // 0 == error, 1-12 ok
                if (!(em = (validate_month(optarg, mmnames)))){
                    printf("Invalid month %s\n", optarg);
                    arglist_print(argc, argv);
                    help(1);
                }
                break;
            case 'c':
                hm = validate_time(optarg);
                if (hm->invalid) {
                    printf("Invalid time %s\n", optarg);
                    arglist_print(argc, argv);
                    help(1);
                }
                amflag = hm->amflag;
                hh = hm->hh;
                mm = hm->mm;
                print_time = 1;
                free(hm);
                break;
            case 'l':
                free(headline); // it got initialised as a default
                headline = strdup(optarg);
                break;
            case 'i':
                inh_header = 1;
                break;
            case 'm':
                mergeable = 1;
                break;
            case 't':
                trunc_dm = 1;
                break;
            case ':':
            case '?':
                printf("Error in options:\n");
                arglist_print(argc, argv);
                if (c == '?')
                    printf("Illegal option supplied\n");
                else
                    printf("Option requires an argument\n");
                help(1);
                break;
            default:
                help(1);
                break;
        }// switch()
    } // while()
    // OK having processed the users inputs; make sure I have everything
    if (diw < 0) {
        printf("option -d is required\n");
        arglist_print(argc, argv);
        help(1); // bye
    }
    if (atoi(dim) <= 0) {
        printf("option -w is required\n");
        arglist_print(argc, argv);
        help(1); // bye
    }

    {// initialise our wanted_mask, presently 0
        int masks[6] ={0, 1, 2, 4, 8, 16};
        int c;
        char *wmp = dim;
        while (*wmp){
            c = *wmp - '0';
            wanted_mask |= masks[c];
            wmp++;
        }
    } // wanted mask
    // other options are optional :-)
    /* the main processing:
    1. get the start time which will be the first desired weekday
        in the month, regardless of whether it's on our list
    2. Set the end time, make it the 1st desired weekday in month
        after ending month.
    3. Generate the linked list of every weekday in every month
        in the period of interest.
    4. Traverse the list to set the pim_mask appropriately. NB the
    last day can be (8|16) if the 4th day is also the last.
    5. Traverse the list again deleting the nodes not required.
    6. Print the results.
    */
    tp->tm_hour = hh;
    tp->tm_min  = mm;
    tp->tm_mday = 1;
    tstart      = mktime(tp);   // it's 1-mmm-yy etc now
    tstart      = set_diw(diw, tstart);
    tp          = localtime_r(&tstart, tp);
    yy          = tp->tm_year + 1900;
    mon         = tp->tm_mon + 1;
    tp->tm_mon  += nm;  // take it into the month after
    tp->tm_mday	+= 7;
    tend        = mktime(tp);
    for (tnow   = tstart; tnow < tend; ) {
        tp          = localtime_r(&tnow, tp);
        dd          = new_dd(tp);   // create the date struct
        newnode     = new_node();   // create a list node for it
        newnode->data = dd;
        (void)ul_append(head, newnode);    // append it
        tp->tm_mday += 7;   // possibly the 33rd day or so
        tnow        = mktime(tp);
    }
    // update our dd
    for (test=head; test->next; test = test->next)
        set_pim(test->next);

    // now go thru and delete the nodes not required
    delete_nodes_unwanted(head, yy, mon, nm, wanted_mask, em);
    if (!(inh_header)){
        if (mergeable)
            printf("000000000000\t%s\n", headline);
        else
            printf("%s\n", headline);
    }

    for (test = head; test->next; test = test->next) {
        print_node(test->next->data, ddnames,
                   mmnames, print_time, mergeable, amflag, trunc_dm, 0);
    }
    // destroy_list() wants the list head and then the destructor
    // for the data. In this case ->data is a struct with all atomic
    // data so free() is all that's needed
    destroy_list(head, free);

    return 0;
}// main()
void help(int err_status){
    printf("NAME\n");
    printf("\twhen - outputs dates of selected days within months\n\n");
    printf("SYNOPSIS\n");
    printf("\twhen [OPTION]...\n\n");
    printf("DESCRIPTION\n");
    printf("\twrites to standard out\n\n");
    printf("\t-h prints this help\n\n");
    printf("\t-s - starting month [yymm | ccyymm]\n");
    printf("\tdefault current month; 20th century dates must be input "
            ": 19yymm\n\n");
    printf("\t-n - number of months; default 3\n\n");
    printf("\t-d - day; [0..6](Sun..Sat) or day name,\n");
    printf("\tonly enough chars to avoid ambiguity required.\n\n");
    printf("\t-w - which day(s) [1..4, l(ast)]\n");
    printf("\t13 first and third, 24 second and fourth\n");
    printf("\tdigit order is unimportant and may be combined with 'l'\n\n");
    printf("\t-e - except month [1..12] or by name,\n");
    printf("\t enough chars to avoid ambiguity only.\n\n");
    printf("\t-l - Legend - the headline over the printed output,\n");
    printf("\tthe default is 'Events'\n\n");
    printf("\t-c - Clock - the time of the event,\n");
    printf("\tas [0..23][:|.][0..59] | [0..12][:|.][0..59][am|pm]\n");
    printf("\t24:00 will be accepted as the last of the day,\n");
    printf("\trather than being == 00:00 of next day.\n");
    printf("\tIf not input, no time will be output\n\n");
    printf("\t-i - Inhibit headline printing.\n");
    printf("\tUse when concatenating successive runs.\n\n");
    printf("\t-m - Mergeable, prefix every line with date "
            "string 'ccyymmddhhnn\\t'.\n");
    printf("\tConcatenate successive runs, sort, then 'cut -f2'\n\n");
    printf("\t-t - Truncate day and month to 3 characters only.\n\n");
    printf("\tAlphabetic input may be UPPER case, "
            "lower case or Mixed.\n\n");
    // 0 if user asks for help, 1 if invoked because of user error
    exit(err_status);
}//help()
DATEDATA *new_dd(TIME *tp){
    DATEDATA *tmp = malloc(sizeof (DATEDATA));
    /* My records will carry ordinary calendar time values
     not the figures used in struct tm */
    if (tmp) {
        tmp->yy     = tp->tm_year + 1900;
        tmp->mon    = tp->tm_mon + 1;
        tmp->dd     = tp->tm_mday;
        tmp->wd     = tp->tm_wday;  // [0..6]
        tmp->hh     = tp->tm_hour;
        tmp->mm     = tp->tm_min;
        tmp->pim_mask = 0;
    } else {
        perror("malloc");
        exit(1);
    }
    return tmp;
}// new_dd()
TIME *get_now(void) {
    time_t tt;
    TIME *ret;
    tt = time(NULL);
    if ((ret = malloc(sizeof(TIME)))) {
        ret = localtime_r( &tt, ret );
    } else {
        perror("malloc");
        exit(1);
    }
    return ret;
}// get_now()
TIME *set_dom(TIME *tp, int dom) {
    time_t tt;
    tp->tm_mday = dom;
    tt = mktime(tp);
    tp = localtime_r( &tt, tp );
    return tp;
}// set_dom()
void arglist_print(int argc, char **argv){
    int i;
    for (i = 1; i < argc; i++)
        printf("%s  ", argv[i]);
    puts("");
}// arglist_print()
int weekday_number(char *dayname, char *ddnames[]) {
    /* legal values are [0..6] Su, M, Tu, W, Th, F, Sa (or longer) */
    int ret, i, l, count;
    dayname = propercase(dayname);
    if ( (l = strlen(dayname)) == 1){
        if (atoi(dayname) == 0) {
            if (dayname[0] == '0')
                return 0;
        } else if ((ret = atoi(dayname)) > 0 && ret < 7)
            return ret;
    }count = 0;
    for(i = 0; i < 7; i++) {
        if ( strncmp(dayname, ddnames[i], l) == 0) {
            ret = i;
            count++;
        }
    }
    if (count != 1)
        ret = -1; // not found 0, found dups 2.
    return ret;
}//weekday_number()
char *propercase(char *dayname) {
    int i, l;
    l = strlen(dayname);
    dayname[0] = toupper(dayname[0]);
    for (i = 1; i < l; i++)
        dayname[i] = tolower(dayname[i]);
    dayname[l] = '\0';
    return dayname;
}// propercase()
int validate_dim(char *dim) {
    int c, i;
    i = 0;
    char tmp[10];
    int j=0;
    //don't care about the order it's been input
    while((c = dim[i])) {
        switch(c) {
            case '1':
            case '2':
            case '3':
            case '4':
                tmp[j++] = dim[i];
                break;
            case 'L':
            case 'l':
                tmp[j++] = '0' + 5;
                break;
            case 'a': //discard
            case 's':
            case 't':
            case 'A':
            case 'S':
            case 'T':
                break;
            default:
                return -1;
        }//switch()
        i++;
    }
    tmp[j] = '\0';
    strcpy(dim, tmp);
    return 0;
}// validate_dim()
int validate_month(char *mo, char **monames) {
    int i, l, count, ret;
    char *tmp;
    int m = atoi(mo);
    if ( m > 0 && m < 13)
        return m;
    tmp = propercase(mo);
    count = 0;
    l = strlen(mo);
    for (i = 1; i < 13; i++) {
        if (strncmp(mo, monames[i], l) == 0) {
            ret = i;
            count++;
        }//if()
    }//for()
    if ( count != 1 )
        return 0;   // 0 = not found, 2 or 3 = ambiguous
    return ret;
}// validate_month()
int comp(DATEDATA *left, DATEDATA *right){
    /* don't want any finer granularity than the day
        because in this app all times hh:mm are
        set the same.
    */
    if (left->yy < right->yy)
        return -1;
    if (left->yy > right->yy)
        return 1;
    if (left->mon < right->mon)
        return -1;
    if (left->mon > right->mon)
        return 1;
    if (left->dd < right->dd)
        return -1;
    if (left->dd > right->dd)
	return 1;
    return 0;
} // comp()
void print_node(DATEDATA *tp, char **ddnames, char **mmnames,
        int reqtime, int mergeable, int ampm, int trunc_dm, int dbg){
    char themon[10];
    char theday[10];
    strcpy(themon, mmnames[tp->mon]);
    strcpy(theday, ddnames[tp->wd]);
    if (trunc_dm) {
        themon[3] = '\0';
        theday[3] = '\0';
    }
    if (mergeable)
        printf("%4d%.2d%.2d%.2d%.2d\t",
               tp->yy,
               tp->mon,
               tp->dd,
               tp->hh,
               tp->mm );
    if (reqtime){
        if (!(ampm)){
            printf("%.2d:%.2d",
                   tp->hh,
                   tp->mm );
        } else {
            char *ap;
            if (tp->hh <= 12){
                ap = "AM";
            } else {
                tp->hh -= 12;
                ap = "PM";
            }
            printf("%2d:%.2d %s",
                   tp->hh,
                   tp->mm,
                   ap );
        }// if else(ampm)
    }
    printf(" %s  %2d-%s-%4d",
           theday,
           tp->dd,themon,
           tp->yy );
    if (dbg) {
        printf("\t%d", (int)tp->pim_mask);
    }
    puts("");   // at last a new line
    return;
}// print_node()
/*
typedef struct dd {
	int yy;			// year - as ccyy
	int mon;		// month [1..12]
	int dd;			// day	[1..28,29,30,31]
	int hh;			// hour	[0-23]
	int mm;			// minute	[0-59]
	int wd;			// week day [0..6]
	int nim;		// number in month [1..5]
	int last;		// last week day in month? [0..1]
} DATEDATA;
*/
HM *validate_time(const char *ts) {
    /* This returns the time in 24 hour format
        and a flag to indicate if user
        specified am or pm. Output time
        will be am|pm if user specified
        that way on input.
    */
    HM *hm;
    char *mp;
    char *buf = strdup(ts);
    char *work = strdup(ts);
    hm = (HM*) malloc(sizeof(HM));
    hm->invalid = 0;
    if (hm && buf && work) {
        hm->amflag = 0;
        if (strstr(buf, "am") ||
            strstr(buf, "AM") ||
            strstr(buf, "pm") ||
            strstr(buf, "PM") ){
            hm->amflag = 1;
        }
    // get hours/ minutes
    if ((mp = strstr(work, ":"))) {
        *mp = '\0';
        mp++;
        hm->hh = atoi(work);
        hm->mm = atoi(mp);
    } else if ((mp = strstr(work, "."))) {
        *mp = '\0';
        mp++;
        hm->hh = atoi(work);
        hm->mm = atoi(mp);
    } else {
        hm->hh = atoi(work);
        hm->mm = 0;
    }
    if (strstr(buf, "PM") || strstr(buf, "pm") ){
        if (hm->hh <= 12)
            hm->hh += 12;
        /* NB I am allowing the Thai syntax
        21.00 pm in this function */
    }
    if (hm->hh == 24) {
        hm->hh = 23;
        hm->mm = 59;
    }
    // validation
    if ( hm->hh < 0 || hm->hh >23 || hm->mm < 0 || hm->mm > 59 )
        hm->invalid = 1;
    } else {
        fprintf(stderr, "Could not get memory\n");
        exit(1);
    }
    return hm;
}// validate_time()
time_t set_diw(int diw, time_t tim){
    TIME *tp = (TIME*) malloc(sizeof(TIME));
    if (diw < 0 || diw > 6) {
        fprintf(stderr, "In set_diw() day in week is out of range\n");
        exit(1);
    }
    if (tp) {
        for(;;) {
            tp = localtime_r(&tim, tp);
            if (tp->tm_wday == diw)
                break;
            tp->tm_mday++;
            tim = mktime(tp);
        }
    } else {
        fprintf(stderr, "Could not get memory\n");
        exit(1);
    }
    free(tp);
    return tim;
}// set_diw()
void set_pim(LNODE *node){
    int masks[6] = {0, 1, 2, 4, 8, 16};
    static int lastmonth, dim;
    int month_next_week;
    DATEDATA *nwd;
    DATEDATA *dd = node->data;
    if (lastmonth != dd->mon) // will be true on 1st use also
        dim = 1;
    else
        dim++;
    dd->pim_mask = masks[dim];
    lastmonth = dd->mon;
    if (dim == 4 && node->next){
        nwd = node->next->data;
        month_next_week = nwd->mon;
    } else {
        return;
    }
    // in a month with 5 dims month_next_week will == lastmonth
    if (month_next_week != lastmonth)
        dd->pim_mask |= masks[5];
    return;
}// set_pim()
int get_month_next_week(DATEDATA *dd) {
    TIME tm;
    time_t mydate;
    tm.tm_year  = dd->yy - 1900;
    tm.tm_mon   = dd->mon - 1;
    tm.tm_mday  = dd->dd + 7;
    tm.tm_hour  = dd->hh;
    tm.tm_min   = dd->mm;
    tm.tm_sec   = 0;
    mydate      = mktime(&tm);
    (void)localtime_r(&mydate, &tm);
    return tm.tm_mon + 1;
}// get_month_next_week()
void delete_nodes_unwanted(LNODE *node, int yy, int mon,
                           int nm, int wanted_mask, int except_mon){
    int test_mask;
    int del, maxyy, maxmon;
    DATEDATA *ddp;
    maxyy = yy + (mon + nm)/12;
    maxmon = (mon + nm) % 12;
    LNODE *test = NULL;
    for(test = node; test->next;){
        del = 0;
        ddp = test->next->data;
        test_mask = wanted_mask & ddp->pim_mask;
        del = ((test_mask == 0)     ||
                (ddp->yy > maxyy)   ||
                (ddp->yy == maxyy && ddp->mon >= maxmon)    ||
                (except_mon == ddp->mon));
        if(del){
            (void)node_delete(node, test->next, (void*)free);
        }else{
            test = test->next;
        }//if-else
    }// for()
    return;
}//delete_nodes_unwanted()
